/* libdf3 - create.c */

/* cf fdopen(3), ftruncate(2) */
#define _POSIX_C_SOURCE 200112L

#include "df3_priv.h"

#include <arpa/inet.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>


static int writeHeader(int const fd, df3Axes const a) {
  endianHeader u = {.v[0] = htons((uint16_t)a.x),
                    .v[1] = htons((uint16_t)a.y),
                    .v[2] = htons((uint16_t)a.z)};
  return DF3_HEADERSIZE == write(fd, &u.raw, DF3_HEADERSIZE) ? 0 : errno;
}


df3Error df3_create_file(df3Handle *df3) {
  /* expect partially initialised handle */
  if (!df3  ||  df3->chcks  ||  df3->ec  ||  df3->dp  ||  df3->readonly)
    return ecBadHandle;
  else if (!df3_check_bpv(df3->bpv))
    return (df3->ec = ecBadDimsBpv);
  else if (!df3_check_axes(df3->axes))
    return (df3->ec = ecBadDimsAxes);
  else if (!df3_check_fname(df3->fname))
    return (df3->ec = ecUnusableFile);

  /* file must not exist */
  off_t fsz = 0;
  if (0 != (df3->syserr = df3ExistFileCheck(df3->fname, &fsz)))
    return (df3->ec = ecSysErr);
  else if (fsz)
    return (df3->ec = ecUnusableFile);

  mode_t const mode = S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH;
  int const fd = open(df3->fname, O_RDWR|O_CREAT|O_EXCL, mode);
  if (-1 == fd) {
    df3->syserr = errno;
    return (df3->ec = ecSysErr);
  }

  fsz = (off_t)(AXMUL(df3->axes) * df3->bpv) + DF3_HEADERSIZE;
  if (-1 == ftruncate(fd, fsz)) {
    df3->syserr = errno;
    goto label_fail;
  }
  else if (0 != (df3->syserr = writeHeader(fd, df3->axes)))
    goto label_fail;
  else if (-1 == lseek(fd, 0, SEEK_SET)) {
    df3->syserr = errno;
    goto label_fail;
  }

  switch (df3->bpv) {
    case 1:  /* fall through if large file */
      if ((off_t)DF3_MMAPLIMIT >= fsz) {
        mmFile *p = malloc(sizeof(mmFile));
        if (!p) {
          df3->syserr = errno;
          goto label_fail;
        }
        p->fd = fd;
        p->n = (size_t)fsz;
        p->dp = mmap(NULL, p->n, PROT_READ|PROT_WRITE, MAP_SHARED, p->fd, 0);
        if (MAP_FAILED == p->dp) {
          df3->syserr = errno;
          free(p);
          goto label_fail;
        }
        df3->dp = p;
        df3->ismapped = true;
        break;
      }
    case 2:
    case 4: {
        FILE *fp = fdopen(fd, "rb+");
        if (!fp) {
          df3->syserr = errno;
          goto label_fail;
        }
        df3->dp = fp;
        df3->ismapped = false;
      }
      break;
  }

  df3->readonly = false;
  df3->chcks = df3CalcChecksum(df3);
  return (df3->ec = ecOk);

label_fail:
  (void)close(fd);
  (void)unlink(df3->fname);
  return (df3->ec = ecSysErr);
}

/* ----------------------------------------------------------------- *
 *  this file is part of 'libdf3', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.    *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.              *
 *  all rights reserved.                                             *
 * ----------------------------------------------------------------- */
